import shutil
import time
from datetime import date
from pathlib import Path

import typer
from feedgen.feed import FeedGenerator
from rich import print

from zona import markdown as zmd
from zona import util
from zona.config import ZonaConfig
from zona.issues import errors, warnings
from zona.layout import Layout, discover_layout
from zona.log import get_logger
from zona.metadata import parse_metadata
from zona.models import Item, ItemType
from zona.templates import Templater

logger = get_logger()


class ZonaBuilder:
    def __init__(
        self,
        cli_root: Path | None = None,
        cli_output: Path | None = None,
        draft: bool = False,
    ):
        logger.debug("Initializing Builder...")
        self.layout: Layout = discover_layout(cli_root, cli_output)
        self.config: ZonaConfig = ZonaConfig.from_file(
            self.layout.root / "config.yml"
        )
        if draft:
            self.config.build.include_drafts = True
        self.items: list[Item] = []
        self.item_map: dict[Path, Item] = {}
        self.fresh: bool = True
        self.post_list: list[Item] = []
        self.building: bool = False

    def __bool__(self) -> bool:
        return self.building

    def _discover(self):
        layout = self.layout
        items: list[Item] = []

        base = layout.root / layout.content
        logger.debug(f"Discovering content in {base}.")
        for path in base.rglob("*"):
            if path.is_file() and not util.should_ignore(
                path, patterns=self.config.ignore, base=base
            ):
                # we only parse markdown files not in static/
                destination = layout.output / path.relative_to(base)
                item = Item(
                    source=path,
                    destination=destination,
                    url=str(destination.relative_to(layout.output)),
                )
                if path.name.endswith(
                    ".md"
                ) and not path.is_relative_to(
                    layout.root / "content" / "static"
                ):
                    logger.debug(f"Pre-processing {path}.")
                    item.metadata, item.content = parse_metadata(
                        path, config=self.config
                    )
                    if item.metadata.ignore or (
                        item.metadata.draft
                        and not self.config.build.include_drafts
                    ):
                        continue
                    if item.metadata.post:
                        item.post = True
                    elif item.metadata.post is None:
                        # check if in posts dir?
                        blog_dir = base / Path(self.config.blog.dir)
                        if item.source.is_relative_to(blog_dir):
                            item.post = True
                    item.type = ItemType.MARKDOWN
                    item.copy = False
                    name = destination.stem
                    if name == "index":
                        item.destination = (
                            item.destination.with_suffix(".html")
                        )
                    else:
                        relative = path.relative_to(base).with_suffix(
                            ""
                        )
                        name = relative.stem
                        item.destination = (
                            layout.output
                            / relative.parent
                            / name
                            / "index.html"
                        )
                    rel_url = item.destination.parent.relative_to(
                        layout.output
                    )
                    item.url = (
                        ""
                        if rel_url == Path(".")
                        else rel_url.as_posix()
                    )
                items.append(item)
        self.items = items

    def generate_feed(self) -> bytes:
        post_list = self._get_post_list()
        config = self.config.feed
        logger.debug("Processing RSS config.")
        if config.link.endswith("/"):
            config.link = config.link[:-2]
        fg = FeedGenerator()
        fg.id(config.link)
        fg.title(config.title)
        author = {
            "name": config.author.name,
            "email": config.author.email,
        }
        fg.author(author)
        fg.link(
            href=f"{config.link}/{config.path}",
            rel="self",
            type="application/rss+xml",
        )
        fg.language(config.language)
        fg.description(config.description)

        logger.debug("Generating RSS entries for posts...")
        for post in post_list:
            assert post.metadata
            fe = fg.add_entry()  # pyright: ignore[reportUnknownVariableType]
            fe.id(f"{config.link}{util.normalize_url(post.url)}")  # pyright: ignore[reportUnknownMemberType]
            fe.link(  # pyright: ignore[reportUnknownMemberType]
                href=f"{config.link}{util.normalize_url(post.url)}"
            )
            fe.title(post.metadata.title)  # pyright: ignore[reportUnknownMemberType]
            fe.author(author)  # pyright: ignore[reportUnknownMemberType]
            desc = post.metadata.description
            fe.description(desc)  # pyright: ignore[reportUnknownMemberType]
            date = post.metadata.date
            fe.pubDate(date)  # pyright: ignore[reportUnknownMemberType]
        out: bytes = fg.rss_str(pretty=True)  # pyright: ignore[reportUnknownVariableType]
        assert isinstance(out, bytes)
        return out

    def _get_post_list(self) -> list[Item]:
        assert self.items
        if not self.post_list:
            # sort according to date
            # descending order
            logger.debug("Generating post list...")
            self.post_list = sorted(
                [item for item in self.items if item.post],
                key=lambda item: item.metadata.date
                if item.metadata
                else date.min,
                reverse=True,
            )
        return self.post_list

    def _build(self):
        post_list = self._get_post_list()
        # number of posts
        posts = len(post_list)
        # link post chronology
        logger.debug("Linking posts chronologically...")
        for i, item in enumerate(post_list):
            # prev: older post
            older = post_list[i + 1] if i + 1 < posts else None
            # next: newer post
            newer = post_list[i - 1] if i > 0 else None
            item.older = older
            item.newer = newer

        templater = Templater(
            config=self.config,
            template_dir=self.layout.templates,
            post_list=post_list,
        )
        logger.debug("Building item path: Item map.")
        self.item_map = {
            item.source.resolve(): item for item in self.items
        }

        # write code highlighting stylesheet
        if self.config.markdown.syntax_highlighting.enabled:
            pygments_style = zmd.get_style_defs(self.config)
            pygments_path = (
                self.layout.output / "static" / "pygments.css"
            )
            util.ensure_parents(pygments_path)
            pygments_path.write_text(pygments_style)
        for item in self.item_map.values():
            dst = item.destination
            # print(item)
            # create parent dirs if needed
            if item.type == ItemType.MARKDOWN:
                assert item.content is not None
                # parse markdown and render as html
                raw_html = zmd.md_to_html(
                    config=self.config,
                    content=item.content,
                    resolve_links=True,
                    source=item.source,
                    layout=self.layout,
                    item_map=self.item_map,
                    metadata=item.metadata,
                )
                # TODO: test this
                rendered = templater.render_item(item, raw_html)
                util.ensure_parents(dst)
                dst.write_text(rendered, encoding="utf-8")
            else:
                if item.copy:
                    util.copy_static_file(item.source, dst)

    def build(self, strict: bool = False, _fake_wait: bool = False):
        if _fake_wait:
            from time import sleep

            sleep(3)
        if not self.fresh:
            # remove stale post list
            self.post_list.clear()
        self.building = True
        with warnings, errors:
            start_time = time.time()
            logger.info("Initiating build process...")
            # clean output if applicable
            if (
                self.config.build.clean_output_dir
                and self.layout.output.is_dir()
            ):
                logger.debug(
                    f"Found stale output in {self.layout.output}, cleaning up..."
                )
                # only remove output dir's children
                # to avoid breaking live preview
                for child in self.layout.output.iterdir():
                    if child.is_file() or child.is_symlink():
                        child.unlink()
                    elif child.is_dir():
                        shutil.rmtree(child)
            if not self.fresh:
                self.layout = self.layout.refresh()
            logger.info("Discovering site content...")
            self._discover()
            logger.info("Building site...")
            self._build()
            if self.config.feed.enabled:
                logger.info("Generating RSS feed...")
                rss = self.generate_feed()
                path = self.layout.output / self.config.feed.path
                util.ensure_parents(path)
                logger.debug(f"Writing RSS feed to {path}...")
                path.write_bytes(rss)
            end_time = time.time()
            duration = (end_time - start_time) * 1000

        print(
            f"{'Build' if self.fresh else 'Rebuild'} took {duration:.1f}ms."
        )
        if warnings or errors:
            print(
                f"{'Failed due to' if strict else 'Generated'} {len(warnings)} warning(s) and {len(errors)} error(s)."
            )
            if strict or errors:
                raise typer.Exit(1)
        self.fresh = False
        self.building = False



from dataclasses import dataclass
from datetime import date, datetime, time, tzinfo
from pathlib import Path

import frontmatter
from dacite.config import Config
from dacite.core import from_dict
from dacite.exceptions import DaciteError
from dateutil import parser as date_parser
from yaml import YAMLError

import zona.util
from zona.config import ZonaConfig


@dataclass
class Metadata:
    title: str
    date: datetime
    description: str
    show_title: bool = True
    show_date: bool = True
    show_nav: bool = True
    style: str | None = "/static/style.css"
    header: bool = True
    footer: bool = True
    template: str | None = None
    post: bool | None = None
    draft: bool = False
    ignore: bool = False
    math: bool = True


def ensure_timezone(dt: datetime, tz: tzinfo) -> datetime:
    if dt.tzinfo is None or dt.utcoffset() is None:
        dt = dt.replace(tzinfo=tz)
    return dt


# TODO: migrate to using datetime, where user can optionall specify
# a time as well. if only date is given, default to time.min
def parse_date(
    raw_date: str | datetime | date | object, tz: tzinfo
) -> datetime:
    if isinstance(raw_date, datetime):
        return ensure_timezone(raw_date, tz)
    elif isinstance(raw_date, date):
        return datetime.combine(raw_date, time.min, tzinfo=tz)
    assert isinstance(raw_date, str)
    dt = date_parser.parse(raw_date)
    return ensure_timezone(dt, tz)


def parse_metadata(
    path: Path, config: ZonaConfig
) -> tuple[Metadata, str]:
    """
    Parses a file and returns parsed Metadata and its content. Defaults
    are applied for missing fields. If there is no metadata, a Metadata
    with default values is returned.

    Raises:
        ValueError: If the metadata block is malformed in any way.
    """
    try:
        post = frontmatter.load(str(path))
    except YAMLError as e:
        raise ValueError(f"YAML frontmatter error in {path}: {e}")
    raw_meta = post.metadata or {}
    defaults = {
        "title": zona.util.filename_to_title(path),
        "date": datetime.fromtimestamp(path.stat().st_ctime),
        "description": config.blog.defaults.description,
    }
    meta = {**defaults, **raw_meta}
    meta["date"] = parse_date(meta.get("date"), config.feed.timezone)
    try:
        metadata = from_dict(
            data_class=Metadata,
            data=meta,
            config=Config(check_types=True, strict=True),
        )
    except DaciteError as e:
        raise ValueError(f"Malformed metadata in {path}: {e}")
    return metadata, post.content



import io
import os
import signal
import sys
import tempfile
import threading
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from pathlib import Path
from threading import Lock
from types import FrameType
from typing import override

from rich import print
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from zona import util
from zona.builder import ZonaBuilder
from zona.log import get_logger
from zona.websockets import WebSocketServer

logger = get_logger()


def make_reload_script(
    host: str, port: int, scroll_tolerance: int
) -> str:
    """Generates the JavaScript that must be injected into HTML pages for the live reloading to work."""
    js = util.get_resource("server/inject.js").contents
    js = util.minify_js(js)
    address = f"ws://{host}:{port}"
    for placeholder, value in (
        ("__SOCKET_ADDRESS__", address),
        ("__SCROLL_TOLERANCE__", scroll_tolerance),
    ):
        if placeholder not in js:
            raise ValueError(
                f"{placeholder} missing from reload script template!"
            )
        js = js.replace(placeholder, str(value))
    return f"<script>{js}</script>"


def make_handler_class(script: str, lock: Lock):
    """Build the live reload handler with the script as an attribute."""

    class CustomHandler(LiveReloadHandler):
        pass

    CustomHandler.script = script
    CustomHandler.lock = lock
    return CustomHandler


class LiveReloadHandler(SimpleHTTPRequestHandler):
    """
    Request handler implementing live reloading.
    All logs are suppressed.
    HTML files have the reload script injected before </body>.
    """

    script: str = ""
    lock: "Lock | None" = None

    @override
    def log_message(self, format, *args):  # type: ignore
        pass

    @override
    def send_head(self):
        # wait for build to finish
        if self.lock:
            self.lock.acquire()
            self.lock.release()

        path = Path(self.translate_path(self.path))
        # check if serving path/index.html
        if path.is_dir():
            index_path = path / "index.html"
            if index_path.is_file():
                path = index_path
        # check if serving html file
        if path.suffix in {".html", ".htm"} and self.script != "":
            try:
                logger.debug(f"Injecting reload script: {path}")
                # read the html
                with open(path, "rb") as f:
                    content = f.read().decode("utf-8")
                # inject script at the end of body
                if r"</body>" in content:
                    content = content.replace(
                        "</body>", self.script + "</body>"
                    )
                else:
                    # if no </body>, add to the end
                    content += self.script
                # reencode, prepare headers, serve file
                encoded = content.encode("utf-8")
                self.send_response(200)
                self.send_header(
                    "Content-type", "text/html; charset=utf-8"
                )
                self.send_header("Content-Length", str(len(encoded)))
                self.end_headers()
                return io.BytesIO(encoded)
            except Exception:
                self.send_error(404, "File not found")
                return None
        return super().send_head()


class QuietHandler(SimpleHTTPRequestHandler):
    """SimpleHTTPRequestHandler with logs suppressed."""

    @override
    def log_message(self, format, *args):  # type: ignore
        pass


class ZonaServer(ThreadingHTTPServer):
    """HTTP server implementing live reloading via a WebSocket server.
    Suppresses BrokenPipeError and ConnectionResetError.
    """

    ws_server: WebSocketServer | None = None

    def set_ws_server(self, ws_server: WebSocketServer):
        self.ws_server = ws_server

    @override
    def handle_error(self, request, client_address):  # type: ignore
        _, exc_value = sys.exc_info()[:2]
        if not isinstance(
            exc_value, (BrokenPipeError, ConnectionResetError)
        ):
            super().handle_error(request, client_address)


class ZonaReloadHandler(FileSystemEventHandler):
    """FileSystemEventHandler that rebuilds the website
    and triggers the browser into refreshing over WebSocket."""

    def __init__(
        self,
        builder: ZonaBuilder,
        output: Path,
        lock: Lock,
        ws_server: WebSocketServer | None,
    ):
        self.builder: ZonaBuilder = builder
        self.output: Path = output.resolve()
        self.ws_server: WebSocketServer | None = ws_server
        self._lock: Lock = lock

    def _trigger_rebuild(self, event: FileSystemEvent):
        if self._should_ignore(event):
            return

        # don't rebuild if a build is already happening
        if not self._lock.acquire(blocking=False):
            return
        try:
            logger.info(f"Modified: {event.src_path}, rebuilding...")
            # rebuild static site
            self.builder.build(_fake_wait=False)
            if self.ws_server:
                # trigger browser refresh
                self.ws_server.notify_all()
        finally:
            self._lock.release()

    def _should_ignore(self, event: FileSystemEvent) -> bool:
        path = Path(str(event.src_path)).resolve()
        # ignore if the output directory has been changed
        # to avoid infinite loop
        return (
            self.output in path.parents
            or path == self.output
            or event.is_directory
        )

    @override
    def on_modified(self, event: FileSystemEvent):
        self._trigger_rebuild(event)

    @override
    def on_created(self, event: FileSystemEvent):
        self._trigger_rebuild(event)


def serve(
    root: Path | None = None,
    output: Path | None = None,
    draft: bool = True,
    host: str = "localhost",
    port: int = 8000,
    user_reload: bool | None = None,
    strict: bool = False,
):
    """Serve preview website with live reload and automatic rebuild."""
    # create temp dir, automatic cleanup
    with tempfile.TemporaryDirectory() as tmp:
        builder = ZonaBuilder(root, Path(tmp), draft)
        config = builder.config
        # initial site build
        builder.build(strict)
        # use discovered paths if none provided
        if output is None:
            output = builder.layout.output
        if root is None:
            root = builder.layout.root

        # create a thread lock:
        # only one build or serve should happen at a time
        lock = Lock()
        # use config value unless overridden by user
        reload = config.server.reload.enabled
        if user_reload is not None:
            reload = user_reload
        if reload:
            print("Live reloading is enabled.")
            # spin up websocket server for live reloading
            ws_port = port + 1
            ws_server = WebSocketServer(host, ws_port)
            ws_server.start()
            # generate reload script for injection
            scroll_tolerance = config.server.reload.scroll_tolerance
            reload_script = make_reload_script(
                host, ws_port, scroll_tolerance
            )
            # generate handler with reload script as attribute
            handler = make_handler_class(reload_script, lock)
        else:
            handler = QuietHandler
            ws_server = None
        # serve the output directory
        os.chdir(output)
        # initialize http server
        httpd = ZonaServer(
            server_address=(host, port), RequestHandlerClass=handler
        )
        # link websocket server
        if ws_server:
            httpd.set_ws_server(ws_server)
        # provide link to user
        print(f"Serving {output} at http://{host}:{port}")
        print("Exit with <c-c>")

        # start server in a thread
        server_thread = threading.Thread(
            target=httpd.serve_forever, daemon=True
        )
        server_thread.start()

        # initialize reload handler
        event_handler = ZonaReloadHandler(
            builder=builder,
            output=output,
            ws_server=ws_server,
            lock=lock,
        )
        observer = Observer()
        observer.schedule(
            event_handler, path=str(root / "content"), recursive=True
        )
        templates = root / "templates"
        if templates.is_dir():
            observer.schedule(
                event_handler,
                path=str(templates),
                recursive=True,
            )
        observer.start()

        # function to shut down gracefully
        def shutdown_handler(_a: int, _b: FrameType | None):
            print("Shutting down...")
            observer.stop()
            httpd.shutdown()

        # register shutdown handler
        signal.signal(signal.SIGINT, shutdown_handler)
        signal.signal(signal.SIGTERM, shutdown_handler)

        # start file change watcher
        observer.join()



import logging
import os
from collections.abc import Mapping
from pathlib import Path
from typing import Any, Literal, cast, override

from rich.logging import RichHandler

_LOGGER_NAME = "zona"
TRACE_LEVEL = 5


class ZonaLogger(logging.Logger):
    def trace(
        self,
        msg: object,
        *args: object,
        exc_info: Any = None,
        stack_info: bool = False,
        stacklevel: int = 1,
        extra: Mapping[str, object] | None = None,
    ) -> None:
        if self.isEnabledFor(TRACE_LEVEL):
            self._log(
                TRACE_LEVEL,
                msg,
                args,
                exc_info,
                extra,
                stack_info,
                stacklevel,
            )


logging.addLevelName(level=TRACE_LEVEL, levelName="TRACE")
logging.setLoggerClass(ZonaLogger)
# addLoggingLevel("TRACE", logging.DEBUG - 5)


class PathStrippingFormatter(logging.Formatter):
    def __init__(
        self,
        fmt: str | None = None,
        datefmt: str | None = None,
        style: Literal["%", "{", "$"] = "%",
        validate: bool = True,
        prefix: Path | None = None,
        *,
        defaults: Mapping[str, Any] | None = None,
    ) -> None:
        super().__init__(
            fmt, datefmt, style, validate, defaults=defaults
        )
        self.prefix: str | None = (
            str(prefix.absolute()) if prefix else None
        )

    def set_prefix(self, prefix: Path):
        if not self.prefix:
            self.prefix = str(prefix.absolute())

    @override
    def format(self, record: logging.LogRecord) -> str:
        msg = super().format(record)
        if self.prefix:
            msg = msg.replace(self.prefix + os.sep, "")
        return msg


def setup_logging(
    level: str = "INFO",
    fmt: str | None = "%(message)s",
):
    logger = logging.getLogger(_LOGGER_NAME)
    logger.setLevel(level.upper())

    logger.propagate = False

    if not logger.handlers:
        handler = RichHandler(
            rich_tracebacks=True,
            show_path=False,
            show_time=False,
            # level=5,
        )
        # handler.setLevel(TRACE_LEVEL)
        handler.setLevel(level.upper())
        formatter = PathStrippingFormatter(prefix=None, fmt=fmt)
        handler.setFormatter(formatter)
        logger.addHandler(handler)


def set_logger_prefix(
    logger: logging.Logger, prefix: Path
) -> logging.Logger:
    for handler in logger.handlers:
        formatter = getattr(handler, "formatter", None)
        if isinstance(formatter, PathStrippingFormatter):
            formatter.set_prefix(prefix)
    return logger


def get_logger(name: str = _LOGGER_NAME) -> ZonaLogger:
    return cast(ZonaLogger, logging.getLogger(name))



import re
from typing import Any, override

from click.core import Context
from typer.core import TyperGroup


class AliasGroup(TyperGroup):
    """
    Helper for defining command aliases like this:
        @app.command("a | action | xyz")
        https://github.com/fastapi/typer/issues/132#issuecomment-2417492805
    """

    _CMD_SPLIT_P: re.Pattern[str] = re.compile(r" ?[,|] ?")

    @override
    def get_command(self, ctx: Context, cmd_name: str | Any):
        cmd_name = self._group_cmd_name(cmd_name)
        return super().get_command(ctx, cmd_name)

    def _group_cmd_name(self, default_name: str | Any):
        for cmd in self.commands.values():
            name = cmd.name
            if name and default_name in self._CMD_SPLIT_P.split(name):
                return name
        return default_name



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from zona.metadata import Metadata


class ItemType(Enum):
    MARKDOWN = "markdown"
    HTML = "html"
    IMAGE = "image"


@dataclass
class Item:
    source: Path
    destination: Path
    url: str  # relative to site root
    metadata: Metadata | None = None  # frontmatter
    content: str | None = None
    type: ItemType | None = None
    copy: bool = True
    post: bool = False
    newer: Item | None = None
    older: Item | None = None


# @dataclass
# class BuildCtx:
#     layout: Layout
#     item_map: dict[Path, Item] = field(default_factory=dict)



import fnmatch
import re
import shutil
import string
import tempfile
import weakref
from importlib import resources
from importlib.resources.abc import Traversable
from pathlib import Path
from shutil import copy2
from typing import Any, NamedTuple, override

from zona.log import get_logger

logger = get_logger()


class TempDir:
    """Temporary directory that cleans up when it's garbage collected."""

    def __init__(self):
        self._tempdir: str = tempfile.mkdtemp()
        self.path: Path = Path(self._tempdir)
        self._finalizer: weakref.finalize[Any, Any] = (
            weakref.finalize(self, shutil.rmtree, self._tempdir)
        )

    def remove(self):
        self._finalizer()

    @property
    def removed(self):
        return not self._finalizer.alive

    @override
    def __repr__(self) -> str:
        return f"<TempDir {self.path}>"


class ZonaResource(NamedTuple):
    name: str
    contents: str

    @override
    def __str__(self) -> str:
        return self.name

    @override
    def __repr__(self) -> str:
        return f"<ZonaResource, name: {self.name}, contents: {self.contents}>"

    def write_with_root(self, root: Path, force: bool = False):
        targ = root / Path(self.name)
        ensure_parents(targ)
        if targ.is_file():
            if not force:
                raise FileExistsError(
                    f"Refusing to overwrite {targ} without permission!"
                )
            else:
                logger.debug("Forced resource overwrite.")
        logger.debug(f"Writing resource {self} to {targ}.")
        targ.write_text(self.contents)
        return True


TemplateMap = dict[str, ZonaResource]


def get_resource(path: str) -> ZonaResource:
    """Load the packaged resource in data/path"""
    file = resources.files("zona").joinpath("data", path)
    if file.is_file():
        return ZonaResource(name=path, contents=file.read_text())
    else:
        raise FileNotFoundError(
            f"{path} is not a valid Zona resource!"
        )


def get_resources(subdir: str) -> list[ZonaResource]:
    """Load the packaged resources in data/subdir"""
    out: list[ZonaResource] = []
    base = resources.files("zona").joinpath("data", subdir)

    def walk(trav: Traversable, prefix: str = ""):
        for item in trav.iterdir():
            path = f"{prefix}{item.name}"
            if item.is_dir():
                walk(item, prefix=f"{path}/")
            else:
                out.append(
                    ZonaResource(
                        name=f"{subdir}/{path}",
                        contents=item.read_text(),
                    )
                )

    walk(base)
    return out


def get_resource_dir(subdir: str) -> Path:
    dir = resources.files("zona").joinpath(f"data/{subdir}")
    with resources.as_file(dir) as path:
        assert isinstance(path, Path)
        assert path.is_dir()
        return path


def ensure_parents(target: Path):
    """Ensure the target's parent directories exist."""
    target.parent.mkdir(parents=True, exist_ok=True)


def copy_static_file(src: Path, dst: Path):
    """Copy a static file from one location to another."""
    ensure_parents(dst)
    copy2(src, dst)


def is_empty(path: Path) -> bool:
    """If given a file, check if it has any non-whitespace content.
    If given a directory, check if it has any children."""
    if path.is_file():
        return path.read_text().strip() == ""
    else:
        return not any(path.iterdir())


def filename_to_title(path: Path) -> str:
    name = path.stem
    words = name.replace("-", " ").replace("_", " ")
    return string.capwords(words)


def normalize_url(url: str) -> str:
    if not url.startswith("/"):
        url = "/" + url
    return url


def should_ignore(
    path: Path, patterns: list[str], base: Path
) -> bool:
    rel_path = path.relative_to(base)
    return any(
        fnmatch.fnmatch(str(rel_path), pattern)
        for pattern in patterns
    )


MINIFY_JS_PATTERN = re.compile(
    r"""
    //.*?$    |
    /\*.*?\*/ |
    \s+
    """,
    re.MULTILINE | re.DOTALL | re.VERBOSE,
)


def minify_js(js: str) -> str:
    """Naively minifies JavaScript by stripping comments and whitespace."""
    return MINIFY_JS_PATTERN.sub(
        # replace whitespace with single space,
        # strip comments
        lambda m: " " if m.group(0).isspace() else "",
        js,
    ).strip()



from dataclasses import asdict, dataclass
from pathlib import Path
from zoneinfo import ZoneInfo

import typer
import yaml

from zona import log, util
from zona.config import ZonaConfig, find_config

logger = log.get_logger()


@dataclass
class Layout:
    root: Path
    content: Path
    templates: Path
    output: Path
    shared_templates: util.TempDir | None
    _validate: bool
    has_user_templates: bool = False

    def refresh(self):
        logger.debug("Refreshing layout...")
        if (
            self.shared_templates
            and not self.shared_templates.removed
        ):
            logger.debug("Removing stale templates tempdir...")
            self.shared_templates.remove()
        return self.__class__.from_input(
            root=self.root,
            output=self.output,
            validate=self._validate,
        )

    @classmethod
    def from_input(
        cls,
        root: Path,
        output: Path | None = None,
        validate: bool = True,
    ) -> "Layout":
        layout = cls(
            root=root.resolve(),
            content=(root / "content").resolve(),
            templates=(root / "templates").resolve(),
            output=(root / "public").resolve()
            if not output
            else output,
            shared_templates=None,
            _validate=validate,
        )
        if validate:
            logger.debug("Validating site layout...")
            if not layout.content.is_dir():
                logger.error("Missing required content directory!")
                raise FileNotFoundError(
                    "Missing required content directory!"
                )
            internal_templates = util.get_resource_dir("templates")
            user_templates = layout.templates
            if not user_templates.is_dir() or util.is_empty(
                user_templates
            ):
                logger.debug("Using default template directory.")
                # use the included defaults
                layout.templates = internal_templates
            else:
                layout.has_user_templates = True
                seen: set[str] = set()
                temp = util.TempDir()
                logger.debug(
                    f"Creating shared template directory at {temp}"
                )
                for f in user_templates.iterdir():
                    if f.is_file():
                        util.copy_static_file(f, temp.path)
                        seen.add(f.name)
                for f in internal_templates.iterdir():
                    if f.is_file() and f.name not in seen:
                        util.copy_static_file(f, temp.path)
                layout.shared_templates = temp
                layout.templates = temp.path

        return layout


def discover_root(cli_root: Path | None) -> Path:
    if cli_root:
        logger.debug("Using user provided site root.")
        root = cli_root
    else:
        logger.debug("Discovering site layout...")
        config = find_config(cli_root)
        if config:
            root = config.parent
        else:
            logger.debug("Using CWD as root.")
            root = Path.cwd()
    logger.debug(f"Determined {root} to be project root.")
    log.set_logger_prefix(logger, root)
    return root


def discover_layout(
    cli_root: Path | None = None, cli_output: Path | None = None
) -> Layout:
    root = discover_root(cli_root)
    return Layout.from_input(root, cli_output)


def initialize_site(root: Path | None = None):
    logger.info("Initializing site.")
    # initialize a new project
    if not root:
        logger.debug("No root provided; using CWD.")
        root = Path.cwd()
    root = root.absolute().resolve()
    config = find_config(root)
    if config is not None:
        ans = typer.confirm(
            text=(
                f"A config file already exists at {config}.\n"
                f"Delete it and restore defaults?"
            )
        )
        if ans:
            logger.debug("Unlinking config file.")
            config.unlink()
    # create requires layout
    logger.debug("Generating layout.")
    layout = Layout.from_input(root=root, validate=False)
    # load template resources
    logger.debug("Loading internal templates.")
    # only write the footer
    templates = [util.get_resource("templates/footer.md")]
    logger.debug("Loading internal static content.")
    static = util.get_resources("content")
    for dir, resources in [
        (layout.root, None),
        (layout.content, static),
        (layout.templates, templates),
    ]:
        if not dir.is_dir():
            logger.debug(f"Creating {dir}.")
            dir.mkdir()
        if resources is not None:
            logger.debug("Writing resources.")
            for r in resources:
                logger.debug(f"Writing {(root / Path(r.name))}.")
                (root / Path(r.name)).write_text(r.contents)

    config_path = layout.root / "config.yml"
    logger.debug("Loading default configuation.")
    config = ZonaConfig()
    logger.debug(f"Writing default configuration to {config_path}.")
    config_dict = asdict(config)
    if "feed" in config_dict and "timezone" in config_dict["feed"]:
        tz: ZoneInfo = config_dict["feed"]["timezone"]
        config_dict["feed"]["timezone"] = tz.key
    with open(config_path, "w") as f:
        yaml.dump(
            config_dict,
            f,
            sort_keys=False,
            default_flow_style=False,
            indent=2,
        )



import re
from functools import cache

# pattern matching likely external urls (missing protocol)
LIKELY_EXTERNAL_PATTERN = r"^(?![./#])(?:[a-zA-Z0-9-]+\.)+[a-zA-Z]{2,}(?::\d+)?(?:/[^\s\\]*)?$"
# matching emails
EMAIL_PATTERN = r"(mailto:)?((?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|\"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*\")@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\[(?:(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9]))\.){3}(?:(2(5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])|[a-z0-9-]*[a-z0-9]:(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)\]))"

# identify comments and whitespace in javascript
MINIFY_JS_PATTERN = r"""
    //.*?$    |
    /\*.*?\*/ |
    \s+
    """


@cache
def likely_external():
    """Returns a compiled regex for identifying likely external links."""
    return re.compile(LIKELY_EXTERNAL_PATTERN)


@cache
def email():
    """Returns a compiled regex for matching email addresses.
    The first group is an optional mailto: prefix."""
    return re.compile(EMAIL_PATTERN)


@cache
def minify_js():
    """Returns a compiled regex for JavaScript minifications."""
    return re.compile(
        MINIFY_JS_PATTERN, re.MULTILINE | re.DOTALL | re.VERBOSE
    )



from dataclasses import dataclass
from pathlib import Path
from urllib.parse import urldefrag

from zona import util
from zona.config import ZonaConfig
from zona.issues import warnings
from zona.layout import Layout
from zona.lazy_regex import email, likely_external
from zona.log import get_logger
from zona.models import Item

logger = get_logger()


@dataclass
class ZonaLink:
    href: str = ""
    target: str | None = None


def handle_external_link(
    link: ZonaLink,
    config: ZonaConfig | None,
    source: Path,
):
    if config and config.markdown.links.external_new_tab:
        logger.trace(
            f"Likely external link: {link.href} in file {source}"
        )
        link.target = "_blank"
        return link
    else:
        return link


def resolve_link(
    href: str,
    source: Path,
    config: ZonaConfig | None,
    layout: Layout,
    item_map: dict[Path, Item],
) -> ZonaLink:
    """Attempts to resolve an internal link."""
    link = ZonaLink(href)
    source = source.resolve()
    href_ = href
    resolved = Path()
    if href.startswith(("https://", "http://")):
        return handle_external_link(link, config, source)
    # strip anchor from href
    href, frag = urldefrag(href)
    if frag and not href:
        link.href = f"#{frag}"
        logger.trace(
            f"Resolved self-anchor found in {source}: {link.href}"
        )
        return link

    cur = Path(href)
    # resolve relative to content root
    if href.startswith("/"):
        resolved = (layout.content / cur.relative_to("/")).resolve()
    # treat as relative link and try to resolve
    else:
        resolved = (source.parent / cur).resolve()
    # check if link is real
    real = False
    for suffix in {".md", ".html"}:
        if resolved.with_suffix(suffix).exists():
            real = True
            resolved = resolved.with_suffix(suffix)
            break
    if not real:
        if not bool(likely_external().match(href)):
            # check if matching email
            m = email().match(href)
            if m:
                if not m.group(1):
                    warnings(
                        f"Link {href} is likely an email, but is missing the mailto: prefix"
                    )
            else:
                warnings(
                    f"Link {link.href} in file {source} points to a file that doesn't exist"
                )
        link = handle_external_link(link, config, source)
        return link

    # only substitute if link points to actual file
    # that isn't the self file
    item = item_map.get(resolved)
    if item:
        href = util.normalize_url(item.url)
        # don't sub if it's already correct
        if href_ != href:
            link.href = f"{href}#{frag}" if frag else href
            logger.debug(
                f"Link in file {source}: {href_} resolved to {link.href}"
            )
    else:
        # TODO: fix incorrect warning for links /foobar
        # where content/foobar.md exists
        warnings(
            f"File {source}: resolved path {resolved} from link {href_} exists but not found in item map"
        )
    return link



from __future__ import annotations

from abc import ABC
from collections.abc import Iterator
from typing import Any, override


class Issue(ABC):
    def __init__(self, message: str):
        self.message: str = message
        self._name: str = "ISSUE"
        self.log_level: str = "debug"

    @override
    def __str__(self) -> str:
        return self.message

    @override
    def __repr__(self) -> str:
        return f"[{self._name}] {self.message}"

    def log(self):
        from zona.log import get_logger

        logger = get_logger()
        log_func = getattr(logger, self.log_level, logger.debug)
        log_func(self.message)


class WarningIssue(Issue):
    def __init__(self, message: str):
        super().__init__(message)
        self._name: str = "WARNING"
        self.log_level: str = "warning"


class ErrorIssue(Issue):
    def __init__(self, message: str):
        super().__init__(message)
        self._name: str = "ERROR"
        self.log_level: str = "error"


class IssueCollector:
    _type: type[Issue] = Issue

    def __init__(self):
        self._issues: list[Issue] = []

    def add(self, issue: Issue):
        self._issues.append(issue)
        issue.log()

    def log_all(self):
        for issue in self._issues:
            issue.log()

    def clear(self):
        self._issues.clear()

    def __call__(self, arg: str | Issue):
        if isinstance(arg, Issue):
            self.add(arg)
        else:
            self.add(self._type(arg))

    def __bool__(self):
        return bool(self._issues)

    def __len__(self):
        return len(self._issues)

    def __iter__(self) -> Iterator[Issue]:
        return iter(self._issues)

    def __getitem__(self, index: int) -> Issue:
        return self._issues[index]

    def __enter__(self):
        self.clear()
        return self

    def __exit__(self, *args: Any):
        if self:
            from zona.log import get_logger

            name = type(self).__name__
            get_logger().debug(
                f"{len(self)} {name.lower()}{'' if len(self) == 1 else 's'} total."
            )

    @override
    def __eq__(self, other: object) -> bool:
        if isinstance(other, IssueCollector):
            return self._issues == other._issues
        return False

    @override
    def __str__(self):
        return f"Warnings({len(self)} issue{'s' if len(self) != 1 else ''})"

    @override
    def __repr__(self):
        return f"<Warnings with {len(self)} issue{'s' if len(self) != 1 else ''}>"

    def __contains__(self, msg: Issue):
        return msg in self._issues

    def filter(self, type_: type[Issue]) -> list[Issue]:
        return [i for i in self._issues if isinstance(i, type_)]


class WarningCollector(IssueCollector):
    _type: type[Issue] = WarningIssue


class ErrorCollector(IssueCollector):
    _type: type[Issue] = ErrorIssue


# global instance
warnings = WarningCollector()
errors = ErrorCollector()
